# Copyright © 2024 Dmitry Pronin.
import numpy as np
from scipy.signal import savgol_filter
from scipy.stats import rankdata

import seaborn as sns
import plotly.express as px
from matplotlib import pyplot as plt


def word_dateplot(wordlist, freqs, dates, mode, window_relation=10, 
                  sensitivity=-1, **plot_args):
    plt.figure(figsize=(10, 5), dpi=150)
    if mode == 'freq':
        ytitle = "Частотность (ipm)"
        freqs = freqs*10**6 # to ipm
    if mode == 'rank':
        ytitle = "Ранг"
        plt.yscale('log')
        plt.gca().invert_yaxis()
    if sensitivity == -1:
        for i in range(len(wordlist)):
            plt.plot_date(dates, freqs[i], fmt='-', label=wordlist[i], **plot_args)
    else:
        window_size = len(freqs[0])//window_relation + 2
        for i in range(len(wordlist)):
            plt.plot_date(dates, savgol_filter(freqs[i], window_size, sensitivity, deriv=0), 
                           fmt='-', label=wordlist[i], **plot_args)
    plt.xticks(rotation=90);
    plt.legend(title='Термин')
    plt.grid()
    plt.ylabel(ytitle);

# only for DAR with shape nx2
def specificity_shift(words, norm_related_freqs, colors=['C0', 'C1'], orientation='horizontal'):
    norm_related_freqs = np.array(norm_related_freqs)
    if orientation=='horizontal':
        #fig, ax = plt.subplots(figsize=(8, len(words)/4+0.5), dpi=150)
        plt.subplots(figsize=(8, len(words)/4+0.5), dpi=150)
        sns.barplot(x=-norm_related_freqs[:, 0], y=words, orient=orientation, dodge=True, color=colors[0])
        sns.barplot(x=norm_related_freqs[:, 1], y=words, orient=orientation, dodge=True, color=colors[1])
        plt.xlabel('Типичность')
        ticks = np.array([-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1])
        plt.xticks(ticks, np.abs(ticks.astype(np.float16)))
        plt.xlim([-1, 1])
        plt.ylabel('')
    if orientation=='vertical':
        #fig, ax = plt.subplots(figsize=(len(words)/4+0.5, 8), dpi=150)
        plt.subplots(figsize=(len(words)/4+0.5, 8), dpi=150)
        sns.barplot(y=-norm_related_freqs[:, 0], x=words, orient=orientation, dodge=True, color=colors[0])
        sns.barplot(y=norm_related_freqs[: ,1], x=words, orient=orientation, dodge=True, color=colors[1])
        plt.ylabel('Типичность')
        ticks = np.array([-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1])
        plt.yticks(ticks, np.abs(ticks.astype(np.float16)))
        plt.ylim([-1, 1])
        plt.xlabel('')
        plt.xticks(rotation=90)
    
# only for DAR with shape nx2
def word_shift(words, divergences, xlabel='Сдвиг'):
    divergences = np.array(divergences)
    plt.figure(figsize=(8, len(words)/4), dpi=150)
    ax = sns.barplot(x=divergences, y=words, color='grey')
    plt.setp(ax.patches, linewidth=0)
    plt.ylabel('')
    plt.yticks([])
    plt.xlabel(xlabel)
    xlim = 1.3*np.abs(divergences).max()
    plt.xlim(-xlim, xlim)
    for patch in np.where(divergences>0)[0]:
        ax.patches[patch].set_facecolor('C0')
    for patch in np.where(divergences<0)[0]:
        ax.patches[patch].set_facecolor('C1')
    ax.bar_label(ax.containers[0], words)
    plt.show()

def word_phase_portrait(related_freqs, period_names=None, eps=10**(-10), window_relation=10, 
                        sensitivity=1, annotation_step=None):
    related_freqs = np.array(related_freqs)
    if period_names == None:
        period_names = [str(i) for i in range(len(related_freqs))]
    if annotation_step == None:
        annotation_step = len(related_freqs)//10
    window_size = len(related_freqs)//window_relation + 2
    savgol = savgol_filter(related_freqs, window_size, sensitivity, deriv=0)
    savgol_deriv = savgol_filter(savgol_filter(related_freqs/(eps+sum(related_freqs)), window_size,
                                                sensitivity, deriv=1), window_size, 1, deriv=0)
    cmap = plt.cm.autumn
    colors = [cmap(i / float(len(related_freqs))) for i in range(len(related_freqs))]
    plt.figure(figsize=(8, 8), dpi=150)
    plt.plot(savgol_deriv, savgol, color='black', zorder=0)
    plt.scatter(savgol_deriv, savgol, color=colors, zorder=1)
    plt.scatter(savgol_deriv[0], savgol[0], color='black', zorder=2)
    for i in range(0, len(period_names), annotation_step):
        plt.annotate(period_names[i], (savgol_deriv[i], savgol[i]))
    plt.xlabel('Сила тренда')
    plt.ylabel('Относительная частота')
    plt.grid()


def corpus_map(words, related_freqs, start=0, stop=None, eps=10**(-10), window_relation=10):
    related_freqs = np.array(related_freqs)
    size = related_freqs.shape[0]
    window_size = related_freqs.shape[1]//window_relation + 2
    # Noise reduction
    savgol_freq = savgol_filter(related_freqs, window_size, 1, deriv=0)
    # Normalization of derivative + noise reduction
    savgol_deriv = savgol_filter(savgol_filter(related_freqs/(eps+np.sum(related_freqs, axis=1).reshape(-1, 1)), 
                                                window_size, 1, deriv=1), window_size, 1, deriv=0)
    # Rank by freq and derivative
    freq_mean = np.mean(savgol_freq[:, start:stop], axis=1)
    savgol_freq_rank = 1+len(freq_mean)-rankdata(freq_mean, axis=0).T
    deriv_mean = np.mean(savgol_deriv[:, start:stop], axis=1)
    savgol_deriv_rank = 1+len(deriv_mean)-rankdata(deriv_mean, axis=0).T
    plot_dict = {'Термин': words, 
             'Ранг частоты': savgol_freq_rank, 
             'Ранг тренда' : savgol_deriv_rank}
    fig = px.scatter(plot_dict, x = 'Ранг тренда', y = 'Ранг частоты', hover_name='Термин')
    fig.update_layout(width=700, height=700, 
                hoverlabel=dict(bgcolor="white"),
                xaxis_autorange="reversed",
                yaxis_autorange="reversed")
    fig.show()


# Copyright © 2024 Dmitry Pronin.
import numpy as np
from scipy.sparse import csr_matrix
from scipy.linalg import orthogonal_procrustes
from scipy.spatial.distance import cosine

from pandas import DataFrame

from sklearn.neighbors import NearestNeighbors
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.utils.extmath import randomized_svd

# Word Embedder
class WE:
    def __init__(self):
        self.X0c = []
        self.vocab = []
        self.unique_0 = []

    def fit(self, string: str,  context_size: int, 
            mode, zero_diag=True, pmi_threshold=0.1, 
            gamma=0.5, eps=10**-10, **vectorizer_args):
        # Support function for pmi matrix
        def pmi(matrix, positive=True):
            col_totals = matrix.sum(axis=0)
            total = col_totals.sum()
            row_totals = matrix.sum(axis=1)
            expected = np.outer(row_totals, col_totals) / (total+eps)
            matrix = matrix / (expected+eps)
            # Silence distracting warnings about log(0):
            with np.errstate(divide='ignore'):
                matrix = np.log(matrix)
            matrix[np.isinf(matrix)] = 0.0
            matrix[np.isnan(matrix)] = 0.0  # log(0) = 0
            if positive:
                matrix[matrix < pmi_threshold] = 0.0
            return matrix
        # Fit function
        # check_model = CountVectorizer(**vectorizer_args)
        # check_matrix = check_model
        word_windows = []
        words = string.split()
        for i in range(len(words)-context_size):
            word_windows.append(' '.join(words[i:i+context_size]))
        count_model = CountVectorizer(**vectorizer_args) # default unigram model
        count_model.fit([string])
        X0 = count_model.transform(word_windows)
        self.X0c = (X0.T * X0)
        self.vocab = count_model.get_feature_names_out()
        # Save these words (their vectors are useless)
        if zero_diag:
            self.X0c.setdiag(0)
        if mode == 'proba':
            self.X0c = self.X0c / (np.sum(self.X0c, axis=0)+eps)
        if mode == 'ppmi':
            self.X0c = pmi(self.X0c)
        if mode == 'svd':
            self.X0c = csr_matrix(pmi(self.X0c))
            # making of truncatedSVD
            U0, Sigma0, _ = randomized_svd(self.X0c, 
                              n_components=300,
                              n_iter=5,
                              random_state=42)
            self.X0c = U0 @ np.diag(Sigma0)**gamma

    def find_similar(self, word, n=10, metric=cosine, return_distance=False):
        idx = np.where(self.vocab==word)[0]
        neigh = NearestNeighbors(metric=metric)
        neigh.fit(self.X0c)
        indexes = neigh.kneighbors([self.X0c[idx].tolist()[0]], n, return_distance=return_distance)[0]
        return self.vocab[indexes]
    
    # only for not-svd mode
    def find_collocations(self, word, n=10):
        idx = np.where(self.vocab==word)[0]
        row = self.X0c[idx].A[0]
        sorted_indices = np.argsort(row)[::-1]
        return self.vocab[sorted_indices[:n]]

# Diachronic Word Embedder
class DiWE:
    def __init__(self):
        self.X0c = []
        self.X1c = []
        self.vocab = []
        self.unique_0 = []
        self.unique_1 = []
        self.divergences = []

    def fit(self, string1: str, string2: str,  context_size: int, 
            mode, zero_diag=True, unique_threshold=30, pmi_threshold=0.1, 
            gamma=0.5, eps=10**-10, **vectorizer_args):
        # Support function for pmi matrix
        def pmi(matrix, positive=True):
            col_totals = matrix.sum(axis=0)
            total = col_totals.sum()
            row_totals = matrix.sum(axis=1)
            expected = np.outer(row_totals, col_totals) / (total+eps)
            matrix = matrix / (expected+eps)
            # Silence distracting warnings about log(0):
            with np.errstate(divide='ignore'):
                matrix = np.log(matrix)
            matrix[np.isinf(matrix)] = 0.0
            matrix[np.isnan(matrix)] = 0.0  # log(0) = 0
            if positive:
                matrix[matrix < pmi_threshold] = 0.0
            return matrix
        # Fit function
        # check_model = CountVectorizer(**vectorizer_args)
        # check_matrix = check_model
        tagret_word_windows = []
        target = [string1, string2]
        for c in target:
            word_windows = []
            words = c.split()
            for i in range(len(words)-context_size):
                word_windows.append(' '.join(words[i:i+context_size]))
            tagret_word_windows.append(word_windows)
        count_model = CountVectorizer(**vectorizer_args) # default unigram model
        count_model.fit(target)
        X0 = count_model.transform(tagret_word_windows[0])
        X1 = count_model.transform(tagret_word_windows[1])
        self.X0c = (X0.T * X0)
        self.X1c = (X1.T * X1)
        self.vocab = count_model.get_feature_names_out()
        # Check near-unique words for each matrix
        diagX0 = np.diag(self.X0c.todense())
        diagX1 = np.diag(self.X1c.todense())
        idx_X0 = diagX1 < unique_threshold
        idx_X1 = diagX0 < unique_threshold
        # Save these words (their vectors are useless)
        self.unique_0 = self.vocab[idx_X0]
        self.unique_1 = self.vocab[idx_X1]
        # Select not union for new vocab
        bool_ = ~((idx_X0)|(idx_X1))
        self.vocab = self.vocab[bool_]
        self.X0c = csr_matrix(self.X0c[np.ix_(bool_, bool_)])
        self.X1c = csr_matrix(self.X1c[np.ix_(bool_, bool_)])
        if zero_diag:
            self.X0c.setdiag(0)
            self.X1c.setdiag(0)
        if mode == 'proba':
            self.X0c = self.X0c / (np.sum(self.X0c, axis=0)+eps)
            self.X1c = self.X1c / (np.sum(self.X1c, axis=0)+eps)
        if mode == 'ppmi':
            self.X0c = pmi(self.X0c)
            self.X1c = pmi(self.X1c)
        if mode == 'svd':
            self.X0c = csr_matrix(pmi(self.X0c))
            self.X1c = csr_matrix(pmi(self.X1c))
            # making of truncatedSVD
            U0, Sigma0, _ = randomized_svd(self.X0c, 
                              n_components=300,
                              n_iter=5,
                              random_state=42)
            self.X0c = U0 @ np.diag(Sigma0)**gamma
            U1, Sigma1, _ = randomized_svd(self.X1c, 
                              n_components=300,
                              n_iter=5,
                              random_state=42)
            self.X1c = U1 @ np.diag(Sigma1)**gamma
            # ortogonal aligning of X0c to X1c for further computations
            R, _ = orthogonal_procrustes(self.X0c, self.X1c)
            self.X0c = self.X0c @ R
            # Select unique words:
            self.X0c[np.abs(self.X0c) < eps] = 0


    def find_similar(self, word, key_space, value_space, n=10, metric=cosine, return_distance=False):
        idx = np.where(self.vocab==word)[0]
        neigh = NearestNeighbors(metric=metric)
        neigh.fit(value_space)
        indexes = neigh.kneighbors([key_space[idx].tolist()[0]], n, return_distance=return_distance)[0]
        return self.vocab[indexes]

    # measure a distance between matrices
    def diverge(self, metric, **metric_args):
        self.divergences = metric(self.X0c, self.X1c, axis=1, **metric_args)

    # wordwise divergence
    # best choice is any wordwise_metric from support file
    # [a, b, c] - [d, e, f] = [dist(a-d), dist(b-e), dist(c-f)]
    def word_diverge(self, word, wordwise_metric, **metric_args):
        if word not in self.vocab:
            raise Exception(f"There is no {word} in the vocabulary!")
        idx = np.where(self.vocab==word)[0]
        inp = np.array([self.X0c[idx].tolist()[0], self.X1c[idx].tolist()[0]]).T
        return wordwise_metric(inp, **metric_args)
    
    def to_pandas(self):
        df = DataFrame(columns=['Word', 'Divergence'])
        df['Word'] = self.vocab
        df['Divergence'] = self.divergences
        return df


# Copyright © 2024 Dmitry Pronin.
import numpy as np
from scipy.stats import entropy
from scipy.stats import rankdata
from sklearn.neighbors import NearestNeighbors
from scipy.linalg import norm
from scipy.spatial import distance_matrix


def find_similar(words, norm_related_freqs, target_freqs, n=10, metric=entropy):
    words = np.array(words)
    neigh = NearestNeighbors(metric=metric)
    neigh.fit(norm_related_freqs)
    indexes = neigh.kneighbors([target_freqs], n, return_distance=False)[0]
    return words[indexes]

# wordwise jsd 
def word_jsd(related_freqs, eps=10**(-10)):
    related_freqs = np.array(related_freqs)
    m = np.mean(related_freqs, axis=1)
    f1 = related_freqs[:, 0]
    f2 = related_freqs[:, 1]
    return np.sign(f2-f1) * (m*np.log2(1/(m+eps)) - 0.5*(f1*np.log2(1/(f1+eps)) + f2*np.log2(1/(f2+eps))))

# wordwise rank turbulence divergence
def word_rtd(related_freqs, alpha=1):
    related_freqs = np.array(related_freqs)
    f1 = related_freqs[:, 0]
    f2 = related_freqs[:, 1]
    # rank data
    x1 = 1+len(f1)-rankdata(f1)
    x1 = x1**-1
    x2 = 1+len(f2)-rankdata(f2)
    x2 = x2**-1
    # basic rank divergence
    divergence_elements = \
        (alpha+1)/alpha * \
        (np.abs(x1**alpha - x2**alpha))**(1/(alpha+1))
    # detect freqs > 0
    indices1 = f1>0
    indices2 = f2>0
    # normalization coeficient computation
    N1 = np.sum(indices1)
    N2 = np.sum(indices2)
    ranks1disjoint = N2 + N1/2;
    r1disjoint = 1/ranks1disjoint;
    ranks2disjoint = N1 + N2/2;
    r2disjoint = 1/ranks2disjoint;
    normalization = \
        (alpha+1)/alpha * \
        np.sum((np.abs(x1**alpha - r2disjoint**alpha))**(1/(alpha+1))) + \
        (alpha+1)/alpha * \
        np.sum((np.abs(r1disjoint**alpha - x2**alpha))**(1/(alpha+1)))
    # directional divergence
    return np.sign(x2-x1) * divergence_elements/normalization

# rankturbulence for distributions
def rankturbulence(f1, f2, axis=0, alpha=1):
    f1 = np.asarray(f1)
    f2 = np.asarray(f2)
    # rank data
    x1 = 1+f1.shape[axis]-rankdata(f1, axis=axis)
    x1 = x1**-1
    x2 = 1+f2.shape[axis]-rankdata(f2, axis=axis)
    x2 = x2**-1
    # basic rank divergence
    divergence_elements = \
        (alpha+1)/alpha * \
        (np.abs(x1**alpha - x2**alpha))**(1/(alpha+1))
    # detect freqs > 0
    indices1 = f1>0
    indices2 = f2>0
    # normalization coefficient computation
    N1 = np.sum(indices1, axis=axis, keepdims=True)
    N2 = np.sum(indices2, axis=axis, keepdims=True)
    ranks1disjoint = N2 + N1/2;
    r1disjoint = 1/ranks1disjoint;
    ranks2disjoint = N1 + N2/2;
    r2disjoint = 1/ranks2disjoint;
    normalization = \
        (alpha+1)/alpha * \
        np.sum((np.abs(x1**alpha - r2disjoint**alpha))**(1/(alpha+1)), axis=axis, keepdims=True) + \
        (alpha+1)/alpha * \
        np.sum((np.abs(r1disjoint**alpha - x2**alpha))**(1/(alpha+1)), axis=axis, keepdims=True)
    # base turbulence divergence
    # final formula with normalization
    return np.sum(divergence_elements/normalization, axis=axis)

def cosine_distance(a, b, axis=1, eps=10**(-10)):
    a = np.array(a)
    b = np.array(b)
    return 1-(np.sum(a*b, axis=axis)/(norm(a, axis=axis)*norm(b, axis=axis)+eps))

# Burrow's delta-distance matrix
def delta_distances(freq_matrix, top_n=300, eps=10**-10):
    # Use the top-n threshold
    freq_matrix = np.array(freq_matrix)[:top_n]
    rows, cols =  freq_matrix.shape
    delta_matrix = np.zeros((cols, cols))
    for row in range(rows):
        delta_matrix += 1/(np.std(freq_matrix[row])+eps) * distance_matrix(freq_matrix[row].reshape(-1, 1), 
                                                     freq_matrix[row].reshape(-1, 1))
    return delta_matrix

# Burrow's wordwise delta-distance
# f1 and f2 are parts of the freq_matrix
def word_delta(f1, f2, freq_matrix, top_n=300, eps=10**-10):
    freq_matrix = np.array(freq_matrix)[:top_n]
    f1 = np.array(f1)[:top_n]
    f2 = np.array(f2)[:top_n]
    # Reshaping for 1-D array
    if len(f1.shape) == 1:
        f1 = f1.reshape(-1, 1)
    if len(f2.shape) == 1:
        f2 = f2.reshape(-1, 1)
    # Calculating mean freqs for each group
    f1 = np.mean(f1, axis=1)
    f2 = np.mean(f2, axis=1)
    return  (f1-f2) / (np.std(freq_matrix, axis=1)+eps)

# Base for Novelity-Transcience analysis
def SNT(sources_dar, space_dar, source_period, metric, window_size=6):
    sources_df = sources_dar.to_pandas()
    space_df = space_dar.to_pandas()
    new_df = sources_df.merge(space_df, on='Word')
    sources_df = new_df[sources_dar.period_names]
    space_df = new_df[space_dar.period_names]
    date_quants = space_dar.period_names #space_df.columns[1:]
    start = source_period[0]
    stop = source_period[-1]
    start_index = list(date_quants).index(start)
    stop_index = list(date_quants).index(stop)
    left_win = start_index-window_size
    right_win = stop_index+window_size+1
    if left_win < 0 or right_win > len(date_quants):
        raise Exception('Invalid source period for selected window size!')
    else:
        dist_matrix = []
        for source in sources_df.columns:
            source_dists = []
            for q in date_quants[left_win:right_win]:
                source_dists.append(metric(space_df[q], sources_df[source]))
            dist_matrix.append(source_dists)
        dist_matrix = np.array(dist_matrix)
        novelity = np.sum(dist_matrix[:, :window_size], axis=1)/window_size
        transience = np.sum(dist_matrix[:, -window_size:], axis=1)/window_size
        return novelity, transience, dist_matrix



# Copyright © 2024 Dmitry Pronin.
from scipy.signal import savgol_filter


def ensemble_filter(data: list, n_filters=100,
                    polyorder=0, **savgol_args) -> list:
    """
    Smooth the timeseries by the ensemble filter which uses
    an averanging of the results of several Savitzky-Golay filters
    with different window sizes

    Parameters
    ----------
    data : array-like
        The timeseries to be filtered
    n_filters : int
        The number of Savitzky-Golay filters in the ensemble
    polyorder : int
        The order of the polynomial used to fit the samples.
        `polyorder` must be less than `window_length`.
    **savgol_args
        Hyperparameters of the Savitzky-Golay filter from scipy
    """
    filt = 0
    start = len(data)//10
    stop = len(data)//4
    step = (stop-start)//n_filters
    if step == 0:
        step = 1
        n_filters = (stop-start)//step
    for _ in range(start, stop, step):
        res = savgol_filter(data, _, polyorder, mode='mirror', **savgol_args)
        filt += res
    return filt/n_filters


def cascade_filter(data: list, n_iter: int, **savgol_args) -> list:
    """
    Smooth the timeseries by the cascade filter which uses a sequential
    Savitzky-Golay filters

    Parameters
    ----------
    data : array-like
        The timeseries to be filtered
    n_iter : int
        The length of the Savitzky-Golay filters sequence
    **savgol_args
        Hyperparameters of the Savitzky-Golay filter from scipy
    """
    for _ in range(n_iter):
        data = savgol_filter(data, mode='mirror', **savgol_args)
    return data



from setuptools import setup, find_packages

setup(
    name="sociolitics",               # Имя вашей библиотеки (уникальное на PyPI)
    version="0.1.5",                 # Версия
    author="Dmitry Pronin",               # Автор
    author_email="DDPronin003@yandex.ru",# Email автора
    description="Sociolit project analysis tools", 
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="", # Ссылка на репозиторий
    packages=find_packages(),        # Автоматический поиск пакетов
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9.7',         # Минимальная версия Python
    install_requires=[

            "pandas>=2.0.2",
            "numpy>=1.21.6",
            "scipy>=1.9.1",
            "scikit-learn>=1.0.2",
            "plotly>=5.9.0",
            "matplotlib>=3.5.2",
            "seaborn>=0.13.0", 
            "joblib>=1.4.2",
            "pymystem3>=0.2.0"

],
)